=== modified file 'bzrlib/_dirstate_helpers_c.pyx'
--- bzrlib/_dirstate_helpers_c.pyx	2008-09-15 03:44:10 +0000
+++ bzrlib/_dirstate_helpers_c.pyx	2008-09-18 01:16:07 +0000
@@ -20,12 +20,24 @@
 """
 
 import binascii
+import bisect
+import errno
 import os
+import stat
 
 from bzrlib import cache_utf8, errors, osutils
 from bzrlib.dirstate import DirState, pack_stat
-from bzrlib.osutils import pathjoin
-
+from bzrlib.osutils import pathjoin, splitpath
+
+
+# This is the Windows equivalent of ENOTDIR
+# It is defined in pywin32.winerror, but we don't want a strong dependency for
+# just an error code.
+# XXX: Perhaps we could get it from a windows header ?
+cdef int ERROR_PATH_NOT_FOUND
+ERROR_PATH_NOT_FOUND = 3
+cdef int ERROR_DIRECTORY
+ERROR_DIRECTORY = 267
 
 # Give Pyrex some function definitions for it to understand.
 # All of these are just hints to Pyrex, so that it can try to convert python
@@ -137,6 +149,7 @@
 
 cdef object safe_string_from_size(char *s, Py_ssize_t size):
     if size < 0:
+        # XXX: On 64-bit machines the <int> cast causes a C compiler warning.
         raise AssertionError(
             'tried to create a string with an invalid size: %d @0x%x'
             % (size, <int>s))
@@ -790,6 +803,7 @@
     """
     return _update_entry(self, entry, abspath, stat_value)
 
+
 cdef _update_entry(self, entry, abspath, stat_value):
     """Update the entry based on what is actually on disk.
 
@@ -935,8 +949,29 @@
     cdef object utf8_decode
     cdef readonly object searched_specific_files
     cdef object search_specific_files
+    cdef object state
+    # Current iteration variables:
+    cdef object current_root
+    cdef object current_root_unicode
+    cdef object root_entries
+    cdef int root_entries_pos, root_entries_len
+    cdef object root_abspath
+    cdef int path_handled
+    cdef int source_index, target_index
+    cdef int want_unversioned
+    cdef object tree
+    cdef object dir_iterator
+    cdef int block_index
+    cdef object current_block
+    cdef int current_block_pos
+    cdef object current_dir_info
+    cdef object root_dir_info
+    cdef int path_index
+    cdef object bisect_left
 
-    def __init__(self, include_unchanged, use_filesystem_for_exec, search_specific_files):
+    def __init__(self, include_unchanged, use_filesystem_for_exec,
+        search_specific_files, state, source_index, target_index,
+        want_unversioned, tree):
         self.old_dirname_to_file_id = {}
         self.new_dirname_to_file_id = {}
         # Just a sentry, so that _process_entry can say that this
@@ -955,8 +990,30 @@
         # detail is not relocated, add the id.
         self.searched_specific_files = set()
         self.search_specific_files = search_specific_files
+        self.state = state
+        self.current_root = None
+        self.current_root_unicode = None
+        self.root_entries = None
+        self.root_entries_pos = 0
+        self.root_entries_len = 0
+        self.root_abspath = None
+        if source_index is None:
+            self.source_index = -1
+        else:
+            self.source_index = source_index
+        self.target_index = target_index
+        self.want_unversioned = want_unversioned
+        self.tree = tree
+        self.dir_iterator = None
+        self.block_index = -1
+        self.current_block = None
+        self.current_block_pos = -1
+        self.current_dir_info = None
+        self.path_index = 0
+        self.root_dir_info = None
+        self.bisect_left = bisect.bisect_left
 
-    def _process_entry(self, entry, path_info, source_index, int target_index, state):
+    def _process_entry(self, entry, path_info):
         """Compare an entry and real disk to generate delta information.
 
         :param path_info: top_relpath, basename, kind, lstat, abspath for
@@ -977,18 +1034,18 @@
         cdef object details_list
         file_id = None
         details_list = entry[1]
-        if source_index is None:
+        if -1 == self.source_index:
             source_details = DirState.NULL_PARENT_DETAILS
         else:
-            source_details = details_list[source_index]
-        target_details = details_list[target_index]
+            source_details = details_list[self.source_index]
+        target_details = details_list[self.target_index]
         target_minikind = _minikind_from_string(target_details[0])
         if path_info is not None and _versioned_minikind(target_minikind):
-            if target_index != 0:
+            if self.target_index != 0:
                 raise AssertionError("Unsupported target index %d" % target_index)
-            link_or_sha1 = _update_entry(state, entry, path_info[4], path_info[3])
+            link_or_sha1 = _update_entry(self.state, entry, path_info[4], path_info[3])
             # The entry may have been modified by update_entry
-            target_details = details_list[target_index]
+            target_details = details_list[self.target_index]
             target_minikind = _minikind_from_string(target_details[0])
         else:
             link_or_sha1 = None
@@ -1017,16 +1074,16 @@
                 old_path = source_details[1]
                 old_dirname, old_basename = os.path.split(old_path)
                 path = pathjoin(entry[0][0], entry[0][1])
-                old_entry = state._get_entry(source_index,
+                old_entry = self.state._get_entry(self.source_index,
                                              path_utf8=old_path)
                 # update the source details variable to be the real
                 # location.
                 if old_entry == (None, None):
-                    raise errors.CorruptDirstate(state._filename,
+                    raise errors.CorruptDirstate(self.state._filename,
                         "entry '%s/%s' is considered renamed from %r"
                         " but source does not exist\n"
                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
-                source_details = old_entry[1][source_index]
+                source_details = old_entry[1][self.source_index]
                 source_minikind = _minikind_from_string(source_details[0])
             if path_info is None:
                 # the file is missing on disk, show as removed.
@@ -1088,7 +1145,7 @@
                 try:
                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
                 except KeyError:
-                    source_parent_entry = state._get_entry(source_index,
+                    source_parent_entry = self.state._get_entry(self.source_index,
                                                            path_utf8=old_dirname)
                     source_parent_id = source_parent_entry[0][2]
                 if source_parent_id == entry[0][2]:
@@ -1106,7 +1163,7 @@
                 except KeyError:
                     # TODO: We don't always need to do the lookup, because the
                     #       parent entry will be the same as the source entry.
-                    target_parent_entry = state._get_entry(target_index,
+                    target_parent_entry = self.state._get_entry(self.target_index,
                                                            path_utf8=new_dirname)
                     if target_parent_entry == (None, None):
                         raise AssertionError(
@@ -1128,7 +1185,8 @@
                 or source_exec != target_exec
                 ):
                 if old_path is None:
-                    old_path = path = pathjoin(old_dirname, old_basename)
+                    path = pathjoin(old_dirname, old_basename)
+                    old_path = path
                     old_path_u = self.utf8_decode(old_path)[0]
                     path_u = old_path_u
                 else:
@@ -1153,7 +1211,7 @@
             path = pathjoin(entry[0][0], entry[0][1])
             # parent id is the entry for the path in the target tree
             # TODO: these are the same for an entire directory: cache em.
-            parent_id = state._get_entry(target_index,
+            parent_id = self.state._get_entry(self.target_index,
                                          path_utf8=entry[0][0])[0][2]
             if parent_id == entry[0][2]:
                 parent_id = None
@@ -1192,7 +1250,7 @@
             # perhaps we should change this - then it would be an unknown.
             old_path = pathjoin(entry[0][0], entry[0][1])
             # parent id is the entry for the path in the target tree
-            parent_id = state._get_entry(source_index, path_utf8=entry[0][0])[0][2]
+            parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
             if parent_id == entry[0][2]:
                 parent_id = None
             return (entry[0][2],
@@ -1222,3 +1280,415 @@
                 % (source_minikind, target_minikind))
             ## import pdb;pdb.set_trace()
         return None
+
+    def __iter__(self):
+        return self
+
+    def iter_changes(self):
+        return self
+
+    cdef void _update_current_block(self):
+        if (self.block_index < len(self.state._dirblocks) and
+            osutils.is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
+            self.current_block = self.state._dirblocks[self.block_index]
+            self.current_block_pos = 0
+        else:
+            self.current_block = None
+
+    def __next__(self):
+        # Simple thunk to allow tail recursion without pyrex confusion
+        return self._iter_next()
+
+    cdef _iter_next(self):
+        """Iterate over the changes."""
+        # This function single steps through an iterator. As such while loops
+        # are often exited by 'return' - the code is structured so that the
+        # next call into the function will return to the same while loop. Note
+        # that all flow control needed to re-reach that step is reexecuted,
+        # which can be a performance problem. It has not yet been tuned to
+        # minimise this; a state machine is probably the simplest restructuring
+        # to both minimise this overhead and make the code considerably more
+        # understandable.
+
+        # sketch: 
+        # compare source_index and target_index at or under each element of search_specific_files.
+        # follow the following comparison table. Note that we only want to do diff operations when
+        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
+        # for the target.
+        # cases:
+        # 
+        # Source | Target | disk | action
+        #   r    | fdlt   |      | add source to search, add id path move and perform
+        #        |        |      | diff check on source-target
+        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
+        #        |        |      | ???
+        #   r    |  a     |      | add source to search
+        #   r    |  a     |  a   | 
+        #   r    |  r     |      | this path is present in a non-examined tree, skip.
+        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
+        #   a    | fdlt   |      | add new id
+        #   a    | fdlt   |  a   | dangling locally added file, skip
+        #   a    |  a     |      | not present in either tree, skip
+        #   a    |  a     |  a   | not present in any tree, skip
+        #   a    |  r     |      | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #   a    |  r     |  a   | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #  fdlt  | fdlt   |      | content in both: diff them
+        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
+        #  fdlt  |  a     |      | unversioned: output deleted id for now
+        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
+        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+
+        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
+        #       keeping a cache of directories that we have seen.
+        cdef object current_dirname, current_blockname
+        cdef char * current_dirname_c, * current_blockname_c
+        _process_entry = self._process_entry
+        uninteresting = self.uninteresting
+        searched_specific_files = self.searched_specific_files
+        # Are we walking a root?
+        while self.root_entries_pos < self.root_entries_len:
+            entry = self.root_entries[self.root_entries_pos]
+            self.root_entries_pos = self.root_entries_pos + 1
+            result = _process_entry(entry, self.root_dir_info)
+            if result is not None and result is not self.uninteresting:
+                return result
+        # Have we finished the prior root, or never started one ?
+        if self.current_root is None:
+            # TODO: the pending list should be lexically sorted?  the
+            # interface doesn't require it.
+            try:
+                self.current_root = self.search_specific_files.pop()
+            except KeyError:
+                raise StopIteration()
+            self.current_root_unicode = self.current_root.decode('utf8')
+            self.searched_specific_files.add(self.current_root)
+            # process the entries for this containing directory: the rest will be
+            # found by their parents recursively.
+            self.root_entries = self.state._entries_for_path(self.current_root)
+            self.root_entries_len = len(self.root_entries)
+            self.root_abspath = self.tree.abspath(self.current_root_unicode)
+            try:
+                root_stat = os.lstat(self.root_abspath)
+            except OSError, e:
+                if e.errno == errno.ENOENT:
+                    # the path does not exist: let _process_entry know that.
+                    self.root_dir_info = None
+                else:
+                    # some other random error: hand it up.
+                    raise
+            else:
+                self.root_dir_info = ('', self.current_root,
+                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
+                    self.root_abspath)
+                if self.root_dir_info[2] == 'directory':
+                    if self.tree._directory_is_tree_reference(
+                        self.current_root_unicode):
+                        self.root_dir_info = self.root_dir_info[:2] + \
+                            ('tree-reference',) + self.root_dir_info[3:]
+            if not self.root_entries and not self.root_dir_info:
+                # this specified path is not present at all, skip it.
+                # (tail recursion, can do a loop once the full structure is
+                # known).
+                return self._iter_next()
+            self.path_handled = 0
+            self.root_entries_pos = 0
+            # XXX Clarity: This loop is duplicated a out the self.current_root
+            # is None guard above: if we return from it, it completes there
+            # (and the following if block cannot trigger because
+            # self.path_handled must be true, so the if block is not
+            # duplicated.
+            while self.root_entries_pos < self.root_entries_len:
+                entry = self.root_entries[self.root_entries_pos]
+                self.root_entries_pos = self.root_entries_pos + 1
+                result = _process_entry(entry, self.root_dir_info)
+                if result is not None:
+                    self.path_handled = -1
+                    if result is not self.uninteresting:
+                        return result
+            # handle unversioned specified paths:
+            if self.want_unversioned and not self.path_handled and self.root_dir_info:
+                new_executable = bool(
+                    stat.S_ISREG(self.root_dir_info[3].st_mode)
+                    and stat.S_IEXEC & self.root_dir_info[3].st_mode)
+                return (None,
+                       (None, self.current_root_unicode),
+                       True,
+                       (False, False),
+                       (None, None),
+                       (None, splitpath(self.current_root_unicode)[-1]),
+                       (None, self.root_dir_info[2]),
+                       (None, new_executable)
+                      )
+            # If we reach here, the outer flow continues, which enters into the
+            # per-root setup logic.
+        if self.current_dir_info is None and self.current_block is None:
+            # setup iteration of this root:
+            if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
+                self.current_dir_info = None
+            else:
+                self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
+                    prefix=self.current_root)
+                self.path_index = 0
+                try:
+                    self.current_dir_info = self.dir_iterator.next()
+                except OSError, e:
+                    # there may be directories in the inventory even though
+                    # this path is not a file on disk: so mark it as end of
+                    # iterator
+                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
+                        self.current_dir_info = None
+                    elif sys.platform == 'win32':
+                        # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
+                        # python 2.5 has e.errno == EINVAL,
+                        #            and e.winerror == ERROR_DIRECTORY
+                        try:
+                            e_winerror = e.winerror
+                        except AttributeError:
+                            e_winerror = None
+                        win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
+                        if (e.errno in win_errors or e_winerror in win_errors):
+                            self.current_dir_info = None
+                        else:
+                            # Will this really raise the right exception ?
+                            raise
+                    else:
+                        raise
+                else:
+                    if self.current_dir_info[0][0] == '':
+                        # remove .bzr from iteration
+                        bzr_index = self.bisect_left(self.current_dir_info[1], ('.bzr',))
+                        if self.current_dir_info[1][bzr_index][0] != '.bzr':
+                            raise AssertionError()
+                        del self.current_dir_info[1][bzr_index]
+            initial_key = (self.current_root, '', '')
+            self.block_index, _ = self.state._find_block_index_from_key(initial_key)
+            if self.block_index == 0:
+                # we have processed the total root already, but because the
+                # initial key matched it we should skip it here.
+                self.block_index = self.block_index + 1
+            self._update_current_block()
+        # walk until both the directory listing and the versioned metadata
+        # are exhausted. 
+        while (self.current_dir_info is not None
+            or self.current_block is not None):
+            if (self.current_dir_info and self.current_block
+                and self.current_dir_info[0][0] != self.current_block[0]):
+                # Work around pyrex broken heuristic - current_dirname has
+                # the same scope as current_dirname_c
+                current_dirname = self.current_dir_info[0][0]
+                current_dirname_c = PyString_AS_STRING_void(
+                    <void *>current_dirname)
+                current_blockname = self.current_block[0]
+                current_blockname_c = PyString_AS_STRING_void(
+                    <void *>current_blockname)
+                # In the python generator we evaluate this if block once per
+                # dir+block; because we reenter in the pyrex version its being
+                # evaluated once per path: we could cache the result before
+                # doing the while loop and probably save time.
+                if _cmp_by_dirs(current_dirname_c,
+                    PyString_Size(current_dirname),
+                    current_blockname_c,
+                    PyString_Size(current_blockname)) < 0:
+                    # filesystem data refers to paths not covered by the
+                    # dirblock.  this has two possibilities:
+                    # A) it is versioned but empty, so there is no block for it
+                    # B) it is not versioned.
+
+                    # if (A) then we need to recurse into it to check for
+                    # new unknown files or directories.
+                    # if (B) then we should ignore it, because we don't
+                    # recurse into unknown directories.
+                    # We are doing a loop
+                    while self.path_index < len(self.current_dir_info[1]):
+                        current_path_info = self.current_dir_info[1][self.path_index]
+                        # dont descend into this unversioned path if it is
+                        # a dir
+                        if current_path_info[2] in ('directory',
+                                                    'tree-reference'):
+                            del self.current_dir_info[1][self.path_index]
+                            self.path_index = self.path_index - 1
+                        self.path_index = self.path_index + 1
+                        if self.want_unversioned:
+                            if current_path_info[2] == 'directory':
+                                if self.tree._directory_is_tree_reference(
+                                    self.utf8_decode(current_path_info[0])[0]):
+                                    current_path_info = current_path_info[:2] + \
+                                        ('tree-reference',) + current_path_info[3:]
+                            new_executable = bool(
+                                stat.S_ISREG(current_path_info[3].st_mode)
+                                and stat.S_IEXEC & current_path_info[3].st_mode)
+                            return (None,
+                                (None, self.utf8_decode(current_path_info[0])[0]),
+                                True,
+                                (False, False),
+                                (None, None),
+                                (None, self.utf8_decode(current_path_info[1])[0]),
+                                (None, current_path_info[2]),
+                                (None, new_executable))
+                    # This dir info has been handled, go to the next
+                    self.path_index = 0
+                    try:
+                        self.current_dir_info = self.dir_iterator.next()
+                    except StopIteration:
+                        self.current_dir_info = None
+                else: #(dircmp > 0)
+                    # We have a dirblock entry for this location, but there
+                    # is no filesystem path for this. This is most likely
+                    # because a directory was removed from the disk.
+                    # We don't have to report the missing directory,
+                    # because that should have already been handled, but we
+                    # need to handle all of the files that are contained
+                    # within.
+                    while self.current_block_pos < len(self.current_block[1]):
+                        current_entry = self.current_block[1][self.current_block_pos]
+                        self.current_block_pos = self.current_block_pos + 1
+                        # entry referring to file not present on disk.
+                        # advance the entry only, after processing.
+                        result = _process_entry(current_entry, None)
+                        if result is not None:
+                            if result is not self.uninteresting:
+                                return result
+                    self.block_index = self.block_index + 1
+                    self._update_current_block()
+                continue # next loop-on-block/dir
+            # current_dir_info and current_block refer to the same directory.
+            advance_entry = True
+            advance_path = True
+            # Assign local variables for current path and entry:
+            if (self.current_block and
+                self.current_block_pos < len(self.current_block[1])):
+                current_entry = self.current_block[1][self.current_block_pos]
+            else:
+                current_entry = None
+            if (self.current_dir_info and
+                self.path_index < len(self.current_dir_info[1])):
+                current_path_info = self.current_dir_info[1][self.path_index]
+                if current_path_info[2] == 'directory':
+                    if self.tree._directory_is_tree_reference(
+                        self.utf8_decode(current_path_info[0])[0]):
+                        current_path_info = current_path_info[:2] + \
+                            ('tree-reference',) + current_path_info[3:]
+            else:
+                current_path_info = None
+            self.path_handled = 0
+            while (current_entry is not None or current_path_info is not None):
+                result = None
+                if current_entry is None:
+                    # unversioned -  the check for path_handled when the path
+                    # is advanced will yield this path if needed.
+                    pass
+                elif current_path_info is None:
+                    # no path is fine: the per entry code will handle it.
+                    result = _process_entry(current_entry, current_path_info)
+                    if result is not None:
+                        if result is self.uninteresting:
+                            result = None
+                elif (current_entry[0][1] != current_path_info[1]
+                      or current_entry[1][self.target_index][0] in 'ar'):
+                    # The current path on disk doesn't match the dirblock
+                    # record. Either the dirblock is marked as absent, or
+                    # the file on disk is not present at all in the
+                    # dirblock. Either way, report about the dirblock
+                    # entry, and let other code handle the filesystem one.
+
+                    # Compare the basename for these files to determine
+                    # which comes first
+                    if current_path_info[1] < current_entry[0][1]:
+                        # extra file on disk: pass for now, but only
+                        # increment the path, not the entry
+                        advance_entry = False
+                    else:
+                        # entry referring to file not present on disk.
+                        # advance the entry only, after processing.
+                        result = _process_entry(current_entry, None)
+                        if result is not None:
+                            if result is self.uninteresting:
+                                result = None
+                        advance_path = False
+                else:
+                    result = _process_entry(current_entry, current_path_info)
+                    if result is not None:
+                        self.path_handled = -1
+                        if result is self.uninteresting:
+                            result = None
+                # >- loop control
+                # >- entry
+                if advance_entry and current_entry is not None:
+                    self.current_block_pos = self.current_block_pos + 1
+                    if self.current_block_pos < len(self.current_block[1]):
+                        current_entry = self.current_block[1][self.current_block_pos]
+                    else:
+                        current_entry = None
+                else:
+                    advance_entry = True # reset the advance flaga
+                # >- path
+                if advance_path and current_path_info is not None:
+                    if not self.path_handled:
+                        # unversioned in all regards
+                        if self.want_unversioned:
+                            new_executable = bool(
+                                stat.S_ISREG(current_path_info[3].st_mode)
+                                and stat.S_IEXEC & current_path_info[3].st_mode)
+                            try:
+                                relpath_unicode = self.utf8_decode(current_path_info[0])[0]
+                            except UnicodeDecodeError:
+                                raise errors.BadFilenameEncoding(
+                                    current_path_info[0], osutils._fs_enc)
+                            if result is not None:
+                                raise AssertionError(
+                                    "result is not None: %r" % result)
+                            result = (None,
+                                (None, relpath_unicode),
+                                True,
+                                (False, False),
+                                (None, None),
+                                (None, self.utf8_decode(current_path_info[1])[0]),
+                                (None, current_path_info[2]),
+                                (None, new_executable))
+                        # dont descend into this unversioned path if it is
+                        # a dir
+                        if current_path_info[2] in ('directory'):
+                            del self.current_dir_info[1][self.path_index]
+                            self.path_index = self.path_index - 1
+                    # dont descend the disk iterator into any tree 
+                    # paths.
+                    if current_path_info[2] == 'tree-reference':
+                        del self.current_dir_info[1][self.path_index]
+                        self.path_index = self.path_index - 1
+                    self.path_index = self.path_index + 1
+                    if self.path_index < len(self.current_dir_info[1]):
+                        current_path_info = self.current_dir_info[1][self.path_index]
+                        if current_path_info[2] == 'directory':
+                            if self.tree._directory_is_tree_reference(
+                                current_path_info[0].decode('utf8')):
+                                current_path_info = current_path_info[:2] + \
+                                    ('tree-reference',) + current_path_info[3:]
+                    else:
+                        current_path_info = None
+                    self.path_handled = 0
+                else:
+                    advance_path = True # reset the advance flag.
+                if result is not None:
+                    # Found a result on this pass, yield it
+                    return result
+            if self.current_block is not None:
+                self.block_index = self.block_index + 1
+                self._update_current_block()
+            if self.current_dir_info is not None:
+                self.path_index = 0
+                try:
+                    self.current_dir_info = self.dir_iterator.next()
+                except StopIteration:
+                    self.current_dir_info = None
+        if len(self.search_specific_files):
+            # More supplied paths to process
+            self.current_root = None
+            return self._iter_next()
+        raise StopIteration()

=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2008-09-15 03:23:53 +0000
+++ bzrlib/dirstate.py	2008-09-18 01:16:07 +0000
@@ -211,6 +211,7 @@
 import time
 import zlib
 
+import bzrlib
 from bzrlib import (
     cache_utf8,
     debug,
@@ -220,7 +221,14 @@
     osutils,
     trace,
     )
-from bzrlib.osutils import pathjoin
+from bzrlib.osutils import pathjoin, splitpath
+
+
+# This is the Windows equivalent of ENOTDIR
+# It is defined in pywin32.winerror, but we don't want a strong dependency for
+# just an error code.
+ERROR_PATH_NOT_FOUND = 3
+ERROR_DIRECTORY = 267
 
 
 # compile the struct compiler we need, so as to only do it once
@@ -1036,6 +1044,24 @@
         self._dirblocks[0] = ('', root_block)
         self._dirblocks[1] = ('', contents_of_root_block)
 
+    def _entries_for_path(self, path):
+        """Return a list with all the entries that match path for all ids."""
+        dirname, basename = os.path.split(path)
+        key = (dirname, basename, '')
+        block_index, present = self._find_block_index_from_key(key)
+        if not present:
+            # the block which should contain path is absent.
+            return []
+        result = []
+        block = self._dirblocks[block_index][1]
+        entry_index, _ = self._find_entry_index(key, block)
+        # we may need to look at multiple entries at this path: walk while the specific_files match.
+        while (entry_index < len(block) and
+            block[entry_index][0][0:2] == key[0:2]):
+            result.append(block[entry_index])
+            entry_index += 1
+        return result
+
     def _entry_to_line(self, entry):
         """Serialize entry to a NULL delimited line ready for _get_output_lines.
 
@@ -2704,7 +2730,7 @@
             raise errors.ObjectNotLocked(self)
 
 
-def py_update_entry(self, entry, abspath, stat_value,
+def _update_entry(self, entry, abspath, stat_value,
                  _stat_to_minikind=DirState._stat_to_minikind,
                  _pack_stat=pack_stat):
     """Update the entry based on what is actually on disk.
@@ -2777,18 +2803,18 @@
     self._dirblock_state = DirState.IN_MEMORY_MODIFIED
     return link_or_sha1
 
-update_entry = py_update_entry
-
 
 class ProcessEntryPython(object):
 
     __slots__ = ["old_dirname_to_file_id", "new_dirname_to_file_id", "uninteresting",
         "last_source_parent", "last_target_parent", "include_unchanged",
         "use_filesystem_for_exec", "utf8_decode", "searched_specific_files",
-        "search_specific_files"]
+        "search_specific_files", "state", "source_index", "target_index",
+        "want_unversioned", "tree"]
 
     def __init__(self, include_unchanged, use_filesystem_for_exec,
-        search_specific_files):
+        search_specific_files, state, source_index, target_index,
+        want_unversioned, tree):
         self.old_dirname_to_file_id = {}
         self.new_dirname_to_file_id = {}
         # Just a sentry, so that _process_entry can say that this
@@ -2807,8 +2833,13 @@
         # detail is not relocated, add the id.
         self.searched_specific_files = set()
         self.search_specific_files = search_specific_files
+        self.state = state
+        self.source_index = source_index
+        self.target_index = target_index
+        self.want_unversioned = want_unversioned
+        self.tree = tree
 
-    def _process_entry(self, entry, path_info, source_index, target_index, state):
+    def _process_entry(self, entry, path_info):
         """Compare an entry and real disk to generate delta information.
 
         :param path_info: top_relpath, basename, kind, lstat, abspath for
@@ -2822,19 +2853,19 @@
                  the object 'uninteresting' if these match, but are
                  basically identical.
         """
-        if source_index is None:
+        if self.source_index is None:
             source_details = DirState.NULL_PARENT_DETAILS
         else:
-            source_details = entry[1][source_index]
-        target_details = entry[1][target_index]
+            source_details = entry[1][self.source_index]
+        target_details = entry[1][self.target_index]
         target_minikind = target_details[0]
         if path_info is not None and target_minikind in 'fdlt':
-            if not (target_index == 0):
+            if not (self.target_index == 0):
                 raise AssertionError()
-            link_or_sha1 = update_entry(state, entry,
+            link_or_sha1 = _update_entry(self.state, entry,
                 abspath=path_info[4], stat_value=path_info[3])
             # The entry may have been modified by update_entry
-            target_details = entry[1][target_index]
+            target_details = entry[1][self.target_index]
             target_minikind = target_details[0]
         else:
             link_or_sha1 = None
@@ -2857,16 +2888,16 @@
                 old_path = source_details[1]
                 old_dirname, old_basename = os.path.split(old_path)
                 path = pathjoin(entry[0][0], entry[0][1])
-                old_entry = state._get_entry(source_index,
+                old_entry = self.state._get_entry(self.source_index,
                                              path_utf8=old_path)
                 # update the source details variable to be the real
                 # location.
                 if old_entry == (None, None):
-                    raise errors.CorruptDirstate(state._filename,
+                    raise errors.CorruptDirstate(self.state._filename,
                         "entry '%s/%s' is considered renamed from %r"
                         " but source does not exist\n"
                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
-                source_details = old_entry[1][source_index]
+                source_details = old_entry[1][self.source_index]
                 source_minikind = source_details[0]
             else:
                 old_dirname = entry[0][0]
@@ -2929,7 +2960,7 @@
                 try:
                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
                 except KeyError:
-                    source_parent_entry = state._get_entry(source_index,
+                    source_parent_entry = self.state._get_entry(self.source_index,
                                                            path_utf8=old_dirname)
                     source_parent_id = source_parent_entry[0][2]
                 if source_parent_id == entry[0][2]:
@@ -2947,7 +2978,7 @@
                 except KeyError:
                     # TODO: We don't always need to do the lookup, because the
                     #       parent entry will be the same as the source entry.
-                    target_parent_entry = state._get_entry(target_index,
+                    target_parent_entry = self.state._get_entry(self.target_index,
                                                            path_utf8=new_dirname)
                     if target_parent_entry == (None, None):
                         raise AssertionError(
@@ -2994,7 +3025,7 @@
             path = pathjoin(entry[0][0], entry[0][1])
             # parent id is the entry for the path in the target tree
             # TODO: these are the same for an entire directory: cache em.
-            parent_id = state._get_entry(target_index,
+            parent_id = self.state._get_entry(self.target_index,
                                          path_utf8=entry[0][0])[0][2]
             if parent_id == entry[0][2]:
                 parent_id = None
@@ -3033,7 +3064,7 @@
             # perhaps we should change this - then it would be an unknown.
             old_path = pathjoin(entry[0][0], entry[0][1])
             # parent id is the entry for the path in the target tree
-            parent_id = state._get_entry(source_index, path_utf8=entry[0][0])[0][2]
+            parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
             if parent_id == entry[0][2]:
                 parent_id = None
             return (entry[0][2],
@@ -3062,6 +3093,342 @@
                 % (source_minikind, target_minikind))
             ## import pdb;pdb.set_trace()
         return None
+
+    def __iter__(self):
+        return self
+
+    def iter_changes(self):
+        """Iterate over the changes."""
+        utf8_decode = cache_utf8._utf8_decode
+        cmp_by_dirs = bzrlib.dirstate.cmp_by_dirs
+        _process_entry = self._process_entry
+        uninteresting = self.uninteresting
+        search_specific_files = self.search_specific_files
+        searched_specific_files = self.searched_specific_files
+        # sketch: 
+        # compare source_index and target_index at or under each element of search_specific_files.
+        # follow the following comparison table. Note that we only want to do diff operations when
+        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
+        # for the target.
+        # cases:
+        # 
+        # Source | Target | disk | action
+        #   r    | fdlt   |      | add source to search, add id path move and perform
+        #        |        |      | diff check on source-target
+        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
+        #        |        |      | ???
+        #   r    |  a     |      | add source to search
+        #   r    |  a     |  a   | 
+        #   r    |  r     |      | this path is present in a non-examined tree, skip.
+        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
+        #   a    | fdlt   |      | add new id
+        #   a    | fdlt   |  a   | dangling locally added file, skip
+        #   a    |  a     |      | not present in either tree, skip
+        #   a    |  a     |  a   | not present in any tree, skip
+        #   a    |  r     |      | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #   a    |  r     |  a   | not present in either tree at this path, skip as it
+        #        |        |      | may not be selected by the users list of paths.
+        #  fdlt  | fdlt   |      | content in both: diff them
+        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
+        #  fdlt  |  a     |      | unversioned: output deleted id for now
+        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
+        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
+        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
+        #        |        |      | this id at the other path.
+
+        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
+        #       keeping a cache of directories that we have seen.
+
+        while search_specific_files:
+            # TODO: the pending list should be lexically sorted?  the
+            # interface doesn't require it.
+            current_root = search_specific_files.pop()
+            current_root_unicode = current_root.decode('utf8')
+            searched_specific_files.add(current_root)
+            # process the entries for this containing directory: the rest will be
+            # found by their parents recursively.
+            root_entries = self.state._entries_for_path(current_root)
+            root_abspath = self.tree.abspath(current_root_unicode)
+            try:
+                root_stat = os.lstat(root_abspath)
+            except OSError, e:
+                if e.errno == errno.ENOENT:
+                    # the path does not exist: let _process_entry know that.
+                    root_dir_info = None
+                else:
+                    # some other random error: hand it up.
+                    raise
+            else:
+                root_dir_info = ('', current_root,
+                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
+                    root_abspath)
+                if root_dir_info[2] == 'directory':
+                    if self.tree._directory_is_tree_reference(
+                        current_root.decode('utf8')):
+                        root_dir_info = root_dir_info[:2] + \
+                            ('tree-reference',) + root_dir_info[3:]
+
+            if not root_entries and not root_dir_info:
+                # this specified path is not present at all, skip it.
+                continue
+            path_handled = False
+            for entry in root_entries:
+                result = _process_entry(entry, root_dir_info)
+                if result is not None:
+                    path_handled = True
+                    if result is not uninteresting:
+                        yield result
+            if self.want_unversioned and not path_handled and root_dir_info:
+                new_executable = bool(
+                    stat.S_ISREG(root_dir_info[3].st_mode)
+                    and stat.S_IEXEC & root_dir_info[3].st_mode)
+                yield (None,
+                       (None, current_root_unicode),
+                       True,
+                       (False, False),
+                       (None, None),
+                       (None, splitpath(current_root_unicode)[-1]),
+                       (None, root_dir_info[2]),
+                       (None, new_executable)
+                      )
+            initial_key = (current_root, '', '')
+            block_index, _ = self.state._find_block_index_from_key(initial_key)
+            if block_index == 0:
+                # we have processed the total root already, but because the
+                # initial key matched it we should skip it here.
+                block_index +=1
+            if root_dir_info and root_dir_info[2] == 'tree-reference':
+                current_dir_info = None
+            else:
+                dir_iterator = osutils._walkdirs_utf8(root_abspath, prefix=current_root)
+                try:
+                    current_dir_info = dir_iterator.next()
+                except OSError, e:
+                    # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
+                    # python 2.5 has e.errno == EINVAL,
+                    #            and e.winerror == ERROR_DIRECTORY
+                    e_winerror = getattr(e, 'winerror', None)
+                    win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
+                    # there may be directories in the inventory even though
+                    # this path is not a file on disk: so mark it as end of
+                    # iterator
+                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
+                        current_dir_info = None
+                    elif (sys.platform == 'win32'
+                          and (e.errno in win_errors
+                               or e_winerror in win_errors)):
+                        current_dir_info = None
+                    else:
+                        raise
+                else:
+                    if current_dir_info[0][0] == '':
+                        # remove .bzr from iteration
+                        bzr_index = bisect_left(current_dir_info[1], ('.bzr',))
+                        if current_dir_info[1][bzr_index][0] != '.bzr':
+                            raise AssertionError()
+                        del current_dir_info[1][bzr_index]
+            # walk until both the directory listing and the versioned metadata
+            # are exhausted. 
+            if (block_index < len(self.state._dirblocks) and
+                osutils.is_inside(current_root, self.state._dirblocks[block_index][0])):
+                current_block = self.state._dirblocks[block_index]
+            else:
+                current_block = None
+            while (current_dir_info is not None or
+                   current_block is not None):
+                if (current_dir_info and current_block
+                    and current_dir_info[0][0] != current_block[0]):
+                    if cmp_by_dirs(current_dir_info[0][0], current_block[0]) < 0:
+                        # filesystem data refers to paths not covered by the dirblock.
+                        # this has two possibilities:
+                        # A) it is versioned but empty, so there is no block for it
+                        # B) it is not versioned.
+
+                        # if (A) then we need to recurse into it to check for
+                        # new unknown files or directories.
+                        # if (B) then we should ignore it, because we don't
+                        # recurse into unknown directories.
+                        path_index = 0
+                        while path_index < len(current_dir_info[1]):
+                                current_path_info = current_dir_info[1][path_index]
+                                if self.want_unversioned:
+                                    if current_path_info[2] == 'directory':
+                                        if self.tree._directory_is_tree_reference(
+                                            current_path_info[0].decode('utf8')):
+                                            current_path_info = current_path_info[:2] + \
+                                                ('tree-reference',) + current_path_info[3:]
+                                    new_executable = bool(
+                                        stat.S_ISREG(current_path_info[3].st_mode)
+                                        and stat.S_IEXEC & current_path_info[3].st_mode)
+                                    yield (None,
+                                        (None, utf8_decode(current_path_info[0])[0]),
+                                        True,
+                                        (False, False),
+                                        (None, None),
+                                        (None, utf8_decode(current_path_info[1])[0]),
+                                        (None, current_path_info[2]),
+                                        (None, new_executable))
+                                # dont descend into this unversioned path if it is
+                                # a dir
+                                if current_path_info[2] in ('directory',
+                                                            'tree-reference'):
+                                    del current_dir_info[1][path_index]
+                                    path_index -= 1
+                                path_index += 1
+
+                        # This dir info has been handled, go to the next
+                        try:
+                            current_dir_info = dir_iterator.next()
+                        except StopIteration:
+                            current_dir_info = None
+                    else:
+                        # We have a dirblock entry for this location, but there
+                        # is no filesystem path for this. This is most likely
+                        # because a directory was removed from the disk.
+                        # We don't have to report the missing directory,
+                        # because that should have already been handled, but we
+                        # need to handle all of the files that are contained
+                        # within.
+                        for current_entry in current_block[1]:
+                            # entry referring to file not present on disk.
+                            # advance the entry only, after processing.
+                            result = _process_entry(current_entry, None)
+                            if result is not None:
+                                if result is not uninteresting:
+                                    yield result
+                        block_index +=1
+                        if (block_index < len(self.state._dirblocks) and
+                            osutils.is_inside(current_root,
+                                              self.state._dirblocks[block_index][0])):
+                            current_block = self.state._dirblocks[block_index]
+                        else:
+                            current_block = None
+                    continue
+                entry_index = 0
+                if current_block and entry_index < len(current_block[1]):
+                    current_entry = current_block[1][entry_index]
+                else:
+                    current_entry = None
+                advance_entry = True
+                path_index = 0
+                if current_dir_info and path_index < len(current_dir_info[1]):
+                    current_path_info = current_dir_info[1][path_index]
+                    if current_path_info[2] == 'directory':
+                        if self.tree._directory_is_tree_reference(
+                            current_path_info[0].decode('utf8')):
+                            current_path_info = current_path_info[:2] + \
+                                ('tree-reference',) + current_path_info[3:]
+                else:
+                    current_path_info = None
+                advance_path = True
+                path_handled = False
+                while (current_entry is not None or
+                    current_path_info is not None):
+                    if current_entry is None:
+                        # the check for path_handled when the path is adnvaced
+                        # will yield this path if needed.
+                        pass
+                    elif current_path_info is None:
+                        # no path is fine: the per entry code will handle it.
+                        result = _process_entry(current_entry, current_path_info)
+                        if result is not None:
+                            if result is not uninteresting:
+                                yield result
+                    elif (current_entry[0][1] != current_path_info[1]
+                          or current_entry[1][self.target_index][0] in 'ar'):
+                        # The current path on disk doesn't match the dirblock
+                        # record. Either the dirblock is marked as absent, or
+                        # the file on disk is not present at all in the
+                        # dirblock. Either way, report about the dirblock
+                        # entry, and let other code handle the filesystem one.
+
+                        # Compare the basename for these files to determine
+                        # which comes first
+                        if current_path_info[1] < current_entry[0][1]:
+                            # extra file on disk: pass for now, but only
+                            # increment the path, not the entry
+                            advance_entry = False
+                        else:
+                            # entry referring to file not present on disk.
+                            # advance the entry only, after processing.
+                            result = _process_entry(current_entry, None)
+                            if result is not None:
+                                if result is not uninteresting:
+                                    yield result
+                            advance_path = False
+                    else:
+                        result = _process_entry(current_entry, current_path_info)
+                        if result is not None:
+                            path_handled = True
+                            if result is not uninteresting:
+                                yield result
+                    if advance_entry and current_entry is not None:
+                        entry_index += 1
+                        if entry_index < len(current_block[1]):
+                            current_entry = current_block[1][entry_index]
+                        else:
+                            current_entry = None
+                    else:
+                        advance_entry = True # reset the advance flaga
+                    if advance_path and current_path_info is not None:
+                        if not path_handled:
+                            # unversioned in all regards
+                            if self.want_unversioned:
+                                new_executable = bool(
+                                    stat.S_ISREG(current_path_info[3].st_mode)
+                                    and stat.S_IEXEC & current_path_info[3].st_mode)
+                                try:
+                                    relpath_unicode = utf8_decode(current_path_info[0])[0]
+                                except UnicodeDecodeError:
+                                    raise errors.BadFilenameEncoding(
+                                        current_path_info[0], osutils._fs_enc)
+                                yield (None,
+                                    (None, relpath_unicode),
+                                    True,
+                                    (False, False),
+                                    (None, None),
+                                    (None, utf8_decode(current_path_info[1])[0]),
+                                    (None, current_path_info[2]),
+                                    (None, new_executable))
+                            # dont descend into this unversioned path if it is
+                            # a dir
+                            if current_path_info[2] in ('directory'):
+                                del current_dir_info[1][path_index]
+                                path_index -= 1
+                        # dont descend the disk iterator into any tree 
+                        # paths.
+                        if current_path_info[2] == 'tree-reference':
+                            del current_dir_info[1][path_index]
+                            path_index -= 1
+                        path_index += 1
+                        if path_index < len(current_dir_info[1]):
+                            current_path_info = current_dir_info[1][path_index]
+                            if current_path_info[2] == 'directory':
+                                if self.tree._directory_is_tree_reference(
+                                    current_path_info[0].decode('utf8')):
+                                    current_path_info = current_path_info[:2] + \
+                                        ('tree-reference',) + current_path_info[3:]
+                        else:
+                            current_path_info = None
+                        path_handled = False
+                    else:
+                        advance_path = True # reset the advance flagg.
+                if current_block is not None:
+                    block_index += 1
+                    if (block_index < len(self.state._dirblocks) and
+                        osutils.is_inside(current_root, self.state._dirblocks[block_index][0])):
+                        current_block = self.state._dirblocks[block_index]
+                    else:
+                        current_block = None
+                if current_dir_info is not None:
+                    try:
+                        current_dir_info = dir_iterator.next()
+                    except StopIteration:
+                        current_dir_info = None
 _process_entry = ProcessEntryPython
 
 
@@ -3073,7 +3440,6 @@
         _bisect_path_left_c as _bisect_path_left,
         _bisect_path_right_c as _bisect_path_right,
         cmp_by_dirs_c as cmp_by_dirs,
-        update_entry as update_entry,
         ProcessEntryC as _process_entry,
         )
 except ImportError:

=== modified file 'bzrlib/tests/intertree_implementations/__init__.py'
--- bzrlib/tests/intertree_implementations/__init__.py	2008-04-28 08:57:45 +0000
+++ bzrlib/tests/intertree_implementations/__init__.py	2008-09-18 01:16:07 +0000
@@ -22,6 +22,7 @@
  - tests/test_workingtree.py
 """
 
+import bzrlib
 import bzrlib.errors as errors
 from bzrlib.transport import get_transport
 from bzrlib.tests import (
@@ -71,11 +72,11 @@
              mutable_trees_to_test_trees)
         """
         result = []
-        for (intertree_class,
+        for (label, intertree_class,
             workingtree_format,
             workingtree_format_to,
             mutable_trees_to_test_trees) in formats:
-            scenario = (intertree_class.__name__, {
+            scenario = (label, {
                 "transport_server":self._transport_server,
                 "transport_readonly_server":self._transport_readonly_server,
                 "bzrdir_format":workingtree_format._matchingbzrdir,
@@ -104,14 +105,33 @@
         ]
     test_intertree_permutations = [
         # test InterTree with two default-format working trees.
-        (InterTree, default_tree_format, default_tree_format,
+        (InterTree.__name__, InterTree, default_tree_format, default_tree_format,
          return_provided_trees)]
     for optimiser in InterTree._optimisers:
-        test_intertree_permutations.append(
-            (optimiser,
-             optimiser._matching_from_tree_format,
-             optimiser._matching_to_tree_format,
-             optimiser._test_mutable_trees_to_test_trees))
+        if optimiser is bzrlib.workingtree_4.InterDirStateTree:
+            # Its a little ugly to be conditional here, but less so than having
+            # the optimiser listed twice.
+            # Add once, compiled version
+            test_intertree_permutations.append(
+                (optimiser.__name__ + "(C)",
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser.make_source_parent_tree_compiled_dirstate))
+            # python version
+            test_intertree_permutations.append(
+                (optimiser.__name__ + "(PY)",
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser.make_source_parent_tree_python_dirstate))
+        else:
+            test_intertree_permutations.append(
+                (optimiser.__name__,
+                 optimiser,
+                 optimiser._matching_from_tree_format,
+                 optimiser._matching_to_tree_format,
+                 optimiser._test_mutable_trees_to_test_trees))
     adapter = InterTreeTestProviderAdapter(
         default_transport,
         # None here will cause a readonly decorator to be created

=== modified file 'bzrlib/tests/test__dirstate_helpers.py'
--- bzrlib/tests/test__dirstate_helpers.py	2008-09-14 12:40:43 +0000
+++ bzrlib/tests/test__dirstate_helpers.py	2008-09-18 01:16:07 +0000
@@ -768,14 +768,6 @@
             from bzrlib._dirstate_helpers_py import _read_dirblocks_py
             self.assertIs(_read_dirblocks_py, dirstate._read_dirblocks)
 
-    def test_update_entry(self):
-        if CompiledDirstateHelpersFeature.available():
-            from bzrlib._dirstate_helpers_c import update_entry
-            self.assertIs(update_entry, dirstate.update_entry)
-        else:
-            from bzrlib.dirstate import py_update_entry
-            self.assertIs(py_update_entry, dirstate.py_update_entry)
-
     def test_process_entry(self):
         if CompiledDirstateHelpersFeature.available():
             from bzrlib._dirstate_helpers_c import ProcessEntryC

=== modified file 'bzrlib/workingtree_4.py'
--- bzrlib/workingtree_4.py	2008-09-15 03:23:53 +0000
+++ bzrlib/workingtree_4.py	2008-09-18 01:16:07 +0000
@@ -98,13 +98,6 @@
 from bzrlib.workingtree import WorkingTree, WorkingTree3, WorkingTreeFormat3
 
 
-# This is the Windows equivalent of ENOTDIR
-# It is defined in pywin32.winerror, but we don't want a strong dependency for
-# just an error code.
-ERROR_PATH_NOT_FOUND = 3
-ERROR_DIRECTORY = 267
-
-
 class WorkingTree4(WorkingTree3):
     """This is the Format 4 working tree.
 
@@ -149,6 +142,8 @@
         self._setup_directory_is_tree_reference()
         self._detect_case_handling()
         self._rules_searcher = None
+        #--- allow tests to select the dirstate iter_changes implementation
+        self._iter_changes = dirstate._process_entry
 
     @needs_tree_write_lock
     def _add(self, files, ids, kinds):
@@ -1810,6 +1805,24 @@
         target.set_parent_ids([revid])
         return target.basis_tree(), target
 
+    @classmethod
+    def make_source_parent_tree_python_dirstate(klass, source, target):
+        result = klass.make_source_parent_tree(source, target)
+        result[1]._iter_changes = dirstate.ProcessEntryPython
+        return result
+
+    @classmethod
+    def make_source_parent_tree_compiled_dirstate(klass, source, target):
+        from bzrlib.tests.test__dirstate_helpers import \
+            CompiledDirstateHelpersFeature
+        if not CompiledDirstateHelpersFeature.available():
+            from bzrlib.tests import UnavailableFeature
+            raise UnavailableFeature(CompiledDirstateHelpersFeature)
+        from bzrlib._dirstate_helpers_c import ProcessEntryC
+        result = klass.make_source_parent_tree(source, target)
+        result[1]._iter_changes = ProcessEntryC
+        return result
+
     _matching_from_tree_format = WorkingTreeFormat4()
     _matching_to_tree_format = WorkingTreeFormat4()
     _test_mutable_trees_to_test_trees = make_source_parent_tree
@@ -1836,19 +1849,14 @@
             output. An unversioned file is defined as one with (False, False)
             for the versioned pair.
         """
-        utf8_decode = cache_utf8._utf8_decode
-        _minikind_to_kind = dirstate.DirState._minikind_to_kind
-        cmp_by_dirs = dirstate.cmp_by_dirs
         # NB: show_status depends on being able to pass in non-versioned files
         # and report them as unknown
         # TODO: handle extra trees in the dirstate.
         if (extra_trees or specific_files == []):
             # we can't fast-path these cases (yet)
-            for f in super(InterDirStateTree, self).iter_changes(
+            return super(InterDirStateTree, self).iter_changes(
                 include_unchanged, specific_files, pb, extra_trees,
-                require_versioned, want_unversioned=want_unversioned):
-                yield f
-            return
+                require_versioned, want_unversioned=want_unversioned)
         parent_ids = self.target.get_parent_ids()
         if not (self.source._revision_id in parent_ids
                 or self.source._revision_id == NULL_REVISION):
@@ -1856,14 +1864,6 @@
                 "revision {%s} is not stored in {%s}, but %s "
                 "can only be used for trees stored in the dirstate"
                 % (self.source._revision_id, self.target, self.iter_changes))
-        update_entry = dirstate.update_entry
-        use_filesystem_for_exec = (sys.platform != 'win32')
-        search_specific_files = set()
-        process_entry = dirstate._process_entry(include_unchanged,
-            use_filesystem_for_exec, search_specific_files)
-        _process_entry = process_entry._process_entry
-        uninteresting = process_entry.uninteresting
-        searched_specific_files = process_entry.searched_specific_files
         target_index = 0
         if self.source._revision_id == NULL_REVISION:
             source_index = None
@@ -1879,37 +1879,22 @@
         if specific_files:
             specific_files_utf8 = set()
             for path in specific_files:
+                # Note, if there are many specific files, using cache_utf8
+                # would be good here.
                 specific_files_utf8.add(path.encode('utf8'))
             specific_files = specific_files_utf8
         else:
             specific_files = set([''])
         # -- specific_files is now a utf8 path set --
+        search_specific_files = set()
         # -- get the state object and prepare it.
         state = self.target.current_dirstate()
         state._read_dirblocks_if_needed()
-        def _entries_for_path(path):
-            """Return a list with all the entries that match path for all ids.
-            """
-            dirname, basename = os.path.split(path)
-            key = (dirname, basename, '')
-            block_index, present = state._find_block_index_from_key(key)
-            if not present:
-                # the block which should contain path is absent.
-                return []
-            result = []
-            block = state._dirblocks[block_index][1]
-            entry_index, _ = state._find_entry_index(key, block)
-            # we may need to look at multiple entries at this path: walk while the specific_files match.
-            while (entry_index < len(block) and
-                block[entry_index][0][0:2] == key[0:2]):
-                result.append(block[entry_index])
-                entry_index += 1
-            return result
         if require_versioned:
             # -- check all supplied paths are versioned in a search tree. --
             all_versioned = True
             for path in specific_files:
-                path_entries = _entries_for_path(path)
+                path_entries = state._entries_for_path(path)
                 if not path_entries:
                     # this specified path is not present at all: error
                     all_versioned = False
@@ -1936,330 +1921,12 @@
             if not osutils.is_inside_any(other_specific_files, path):
                 # this is a top level path, we must check it.
                 search_specific_files.add(path)
-        # sketch: 
-        # compare source_index and target_index at or under each element of search_specific_files.
-        # follow the following comparison table. Note that we only want to do diff operations when
-        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
-        # for the target.
-        # cases:
-        # 
-        # Source | Target | disk | action
-        #   r    | fdlt   |      | add source to search, add id path move and perform
-        #        |        |      | diff check on source-target
-        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
-        #        |        |      | ???
-        #   r    |  a     |      | add source to search
-        #   r    |  a     |  a   | 
-        #   r    |  r     |      | this path is present in a non-examined tree, skip.
-        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
-        #   a    | fdlt   |      | add new id
-        #   a    | fdlt   |  a   | dangling locally added file, skip
-        #   a    |  a     |      | not present in either tree, skip
-        #   a    |  a     |  a   | not present in any tree, skip
-        #   a    |  r     |      | not present in either tree at this path, skip as it
-        #        |        |      | may not be selected by the users list of paths.
-        #   a    |  r     |  a   | not present in either tree at this path, skip as it
-        #        |        |      | may not be selected by the users list of paths.
-        #  fdlt  | fdlt   |      | content in both: diff them
-        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
-        #  fdlt  |  a     |      | unversioned: output deleted id for now
-        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
-        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
-        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
-        #        |        |      | this id at the other path.
-        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
-        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
-        #        |        |      | this id at the other path.
-
-        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
-        #       keeping a cache of directories that we have seen.
-
-        while search_specific_files:
-            # TODO: the pending list should be lexically sorted?  the
-            # interface doesn't require it.
-            current_root = search_specific_files.pop()
-            current_root_unicode = current_root.decode('utf8')
-            searched_specific_files.add(current_root)
-            # process the entries for this containing directory: the rest will be
-            # found by their parents recursively.
-            root_entries = _entries_for_path(current_root)
-            root_abspath = self.target.abspath(current_root_unicode)
-            try:
-                root_stat = os.lstat(root_abspath)
-            except OSError, e:
-                if e.errno == errno.ENOENT:
-                    # the path does not exist: let _process_entry know that.
-                    root_dir_info = None
-                else:
-                    # some other random error: hand it up.
-                    raise
-            else:
-                root_dir_info = ('', current_root,
-                    osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
-                    root_abspath)
-                if root_dir_info[2] == 'directory':
-                    if self.target._directory_is_tree_reference(
-                        current_root.decode('utf8')):
-                        root_dir_info = root_dir_info[:2] + \
-                            ('tree-reference',) + root_dir_info[3:]
-
-            if not root_entries and not root_dir_info:
-                # this specified path is not present at all, skip it.
-                continue
-            path_handled = False
-            for entry in root_entries:
-                result = _process_entry(entry, root_dir_info, source_index, target_index, state)
-                if result is not None:
-                    path_handled = True
-                    if result is not uninteresting:
-                        yield result
-            if want_unversioned and not path_handled and root_dir_info:
-                new_executable = bool(
-                    stat.S_ISREG(root_dir_info[3].st_mode)
-                    and stat.S_IEXEC & root_dir_info[3].st_mode)
-                yield (None,
-                       (None, current_root_unicode),
-                       True,
-                       (False, False),
-                       (None, None),
-                       (None, splitpath(current_root_unicode)[-1]),
-                       (None, root_dir_info[2]),
-                       (None, new_executable)
-                      )
-            initial_key = (current_root, '', '')
-            block_index, _ = state._find_block_index_from_key(initial_key)
-            if block_index == 0:
-                # we have processed the total root already, but because the
-                # initial key matched it we should skip it here.
-                block_index +=1
-            if root_dir_info and root_dir_info[2] == 'tree-reference':
-                current_dir_info = None
-            else:
-                dir_iterator = osutils._walkdirs_utf8(root_abspath, prefix=current_root)
-                try:
-                    current_dir_info = dir_iterator.next()
-                except OSError, e:
-                    # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
-                    # python 2.5 has e.errno == EINVAL,
-                    #            and e.winerror == ERROR_DIRECTORY
-                    e_winerror = getattr(e, 'winerror', None)
-                    win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
-                    # there may be directories in the inventory even though
-                    # this path is not a file on disk: so mark it as end of
-                    # iterator
-                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
-                        current_dir_info = None
-                    elif (sys.platform == 'win32'
-                          and (e.errno in win_errors
-                               or e_winerror in win_errors)):
-                        current_dir_info = None
-                    else:
-                        raise
-                else:
-                    if current_dir_info[0][0] == '':
-                        # remove .bzr from iteration
-                        bzr_index = bisect_left(current_dir_info[1], ('.bzr',))
-                        if current_dir_info[1][bzr_index][0] != '.bzr':
-                            raise AssertionError()
-                        del current_dir_info[1][bzr_index]
-            # walk until both the directory listing and the versioned metadata
-            # are exhausted. 
-            if (block_index < len(state._dirblocks) and
-                osutils.is_inside(current_root, state._dirblocks[block_index][0])):
-                current_block = state._dirblocks[block_index]
-            else:
-                current_block = None
-            while (current_dir_info is not None or
-                   current_block is not None):
-                if (current_dir_info and current_block
-                    and current_dir_info[0][0] != current_block[0]):
-                    if cmp_by_dirs(current_dir_info[0][0], current_block[0]) < 0:
-                        # filesystem data refers to paths not covered by the dirblock.
-                        # this has two possibilities:
-                        # A) it is versioned but empty, so there is no block for it
-                        # B) it is not versioned.
-
-                        # if (A) then we need to recurse into it to check for
-                        # new unknown files or directories.
-                        # if (B) then we should ignore it, because we don't
-                        # recurse into unknown directories.
-                        path_index = 0
-                        while path_index < len(current_dir_info[1]):
-                                current_path_info = current_dir_info[1][path_index]
-                                if want_unversioned:
-                                    if current_path_info[2] == 'directory':
-                                        if self.target._directory_is_tree_reference(
-                                            current_path_info[0].decode('utf8')):
-                                            current_path_info = current_path_info[:2] + \
-                                                ('tree-reference',) + current_path_info[3:]
-                                    new_executable = bool(
-                                        stat.S_ISREG(current_path_info[3].st_mode)
-                                        and stat.S_IEXEC & current_path_info[3].st_mode)
-                                    yield (None,
-                                        (None, utf8_decode(current_path_info[0])[0]),
-                                        True,
-                                        (False, False),
-                                        (None, None),
-                                        (None, utf8_decode(current_path_info[1])[0]),
-                                        (None, current_path_info[2]),
-                                        (None, new_executable))
-                                # dont descend into this unversioned path if it is
-                                # a dir
-                                if current_path_info[2] in ('directory',
-                                                            'tree-reference'):
-                                    del current_dir_info[1][path_index]
-                                    path_index -= 1
-                                path_index += 1
-
-                        # This dir info has been handled, go to the next
-                        try:
-                            current_dir_info = dir_iterator.next()
-                        except StopIteration:
-                            current_dir_info = None
-                    else:
-                        # We have a dirblock entry for this location, but there
-                        # is no filesystem path for this. This is most likely
-                        # because a directory was removed from the disk.
-                        # We don't have to report the missing directory,
-                        # because that should have already been handled, but we
-                        # need to handle all of the files that are contained
-                        # within.
-                        for current_entry in current_block[1]:
-                            # entry referring to file not present on disk.
-                            # advance the entry only, after processing.
-                            result = _process_entry(current_entry, None, source_index, target_index, state)
-                            if result is not None:
-                                if result is not uninteresting:
-                                    yield result
-                        block_index +=1
-                        if (block_index < len(state._dirblocks) and
-                            osutils.is_inside(current_root,
-                                              state._dirblocks[block_index][0])):
-                            current_block = state._dirblocks[block_index]
-                        else:
-                            current_block = None
-                    continue
-                entry_index = 0
-                if current_block and entry_index < len(current_block[1]):
-                    current_entry = current_block[1][entry_index]
-                else:
-                    current_entry = None
-                advance_entry = True
-                path_index = 0
-                if current_dir_info and path_index < len(current_dir_info[1]):
-                    current_path_info = current_dir_info[1][path_index]
-                    if current_path_info[2] == 'directory':
-                        if self.target._directory_is_tree_reference(
-                            current_path_info[0].decode('utf8')):
-                            current_path_info = current_path_info[:2] + \
-                                ('tree-reference',) + current_path_info[3:]
-                else:
-                    current_path_info = None
-                advance_path = True
-                path_handled = False
-                while (current_entry is not None or
-                    current_path_info is not None):
-                    if current_entry is None:
-                        # the check for path_handled when the path is adnvaced
-                        # will yield this path if needed.
-                        pass
-                    elif current_path_info is None:
-                        # no path is fine: the per entry code will handle it.
-                        result = _process_entry(current_entry, current_path_info, source_index, target_index, state)
-                        if result is not None:
-                            if result is not uninteresting:
-                                yield result
-                    elif (current_entry[0][1] != current_path_info[1]
-                          or current_entry[1][target_index][0] in 'ar'):
-                        # The current path on disk doesn't match the dirblock
-                        # record. Either the dirblock is marked as absent, or
-                        # the file on disk is not present at all in the
-                        # dirblock. Either way, report about the dirblock
-                        # entry, and let other code handle the filesystem one.
-
-                        # Compare the basename for these files to determine
-                        # which comes first
-                        if current_path_info[1] < current_entry[0][1]:
-                            # extra file on disk: pass for now, but only
-                            # increment the path, not the entry
-                            advance_entry = False
-                        else:
-                            # entry referring to file not present on disk.
-                            # advance the entry only, after processing.
-                            result = _process_entry(current_entry, None, source_index, target_index, state)
-                            if result is not None:
-                                if result is not uninteresting:
-                                    yield result
-                            advance_path = False
-                    else:
-                        result = _process_entry(current_entry, current_path_info, source_index, target_index, state)
-                        if result is not None:
-                            path_handled = True
-                            if result is not uninteresting:
-                                yield result
-                    if advance_entry and current_entry is not None:
-                        entry_index += 1
-                        if entry_index < len(current_block[1]):
-                            current_entry = current_block[1][entry_index]
-                        else:
-                            current_entry = None
-                    else:
-                        advance_entry = True # reset the advance flaga
-                    if advance_path and current_path_info is not None:
-                        if not path_handled:
-                            # unversioned in all regards
-                            if want_unversioned:
-                                new_executable = bool(
-                                    stat.S_ISREG(current_path_info[3].st_mode)
-                                    and stat.S_IEXEC & current_path_info[3].st_mode)
-                                try:
-                                    relpath_unicode = utf8_decode(current_path_info[0])[0]
-                                except UnicodeDecodeError:
-                                    raise errors.BadFilenameEncoding(
-                                        current_path_info[0], osutils._fs_enc)
-                                yield (None,
-                                    (None, relpath_unicode),
-                                    True,
-                                    (False, False),
-                                    (None, None),
-                                    (None, utf8_decode(current_path_info[1])[0]),
-                                    (None, current_path_info[2]),
-                                    (None, new_executable))
-                            # dont descend into this unversioned path if it is
-                            # a dir
-                            if current_path_info[2] in ('directory'):
-                                del current_dir_info[1][path_index]
-                                path_index -= 1
-                        # dont descend the disk iterator into any tree 
-                        # paths.
-                        if current_path_info[2] == 'tree-reference':
-                            del current_dir_info[1][path_index]
-                            path_index -= 1
-                        path_index += 1
-                        if path_index < len(current_dir_info[1]):
-                            current_path_info = current_dir_info[1][path_index]
-                            if current_path_info[2] == 'directory':
-                                if self.target._directory_is_tree_reference(
-                                    current_path_info[0].decode('utf8')):
-                                    current_path_info = current_path_info[:2] + \
-                                        ('tree-reference',) + current_path_info[3:]
-                        else:
-                            current_path_info = None
-                        path_handled = False
-                    else:
-                        advance_path = True # reset the advance flagg.
-                if current_block is not None:
-                    block_index += 1
-                    if (block_index < len(state._dirblocks) and
-                        osutils.is_inside(current_root, state._dirblocks[block_index][0])):
-                        current_block = state._dirblocks[block_index]
-                    else:
-                        current_block = None
-                if current_dir_info is not None:
-                    try:
-                        current_dir_info = dir_iterator.next()
-                    except StopIteration:
-                        current_dir_info = None
+
+        use_filesystem_for_exec = (sys.platform != 'win32')
+        iter_changes = self.target._iter_changes(include_unchanged,
+            use_filesystem_for_exec, search_specific_files, state,
+            source_index, target_index, want_unversioned, self.target)
+        return iter_changes.iter_changes()
 
     @staticmethod
     def is_compatible(source, target):

